qualityoflife = qualityoflife or {};
local q = qualityoflife;

local frameCache = {};
local FRAME_REF = "FRAME_REF";
local FRAME_ORIGINAL_POINTS = "FRAME_ORIGINAL_POINTS";

local TRANSITION_IN_PROGRESS = "TRANSITION_IN_PROGRESS"
local TRANSITION_CANCELLED = "TRANSITION_CANCELLED"
local TRANSITION_TARGET_LEFT = "TRANSITION_TARGET_LEFT";
local TRANSITION_TARGET_BOTTOM = "TRANSITION_TARGET_BOTTOM";
local TRANSITION_SPEED = "TRANSITION_SPEED";
local FINISH_FUNC = "FINISH_FUNC";

local maxD, defaultSpeed = 2, 10;

local function CheckFrameName(frame)
    return frame:GetName() and frame:GetName() ~= "";
end

local function AddFrame(frame)
    if not frameCache[frame:GetName()] then
        frameCache[frame:GetName()] = {};
        frameCache[frame:GetName()][FRAME_REF] = frame;
        frameCache[frame:GetName()][TRANSITION_CANCELLED] = false;
        frameCache[frame:GetName()][FRAME_ORIGINAL_POINTS] = {};
        for i = 1, frame:GetNumPoints() do
            local point, relativeTo, relativePoint, ox, oy = frame:GetPoint(i);
            tinsert(frameCache[frame:GetName()][FRAME_ORIGINAL_POINTS], { point, relativeTo, relativePoint, ox, oy });
        end
    end
    return frameCache[frame:GetName()];
end

function q:ResetFrame(frame)
    if frameCache[frame:GetName()] then
        frame:ClearAllPoints();
        for i, point in ipairs(frameCache[frame:GetName()][FRAME_ORIGINAL_POINTS]) do
            frame:SetPoint(unpack(point));
        end
        frameCache[frame:GetName()] = nil;
    end
end

function q:CancelTransition(frame)
    if frameCache[frame:GetName()] then
        frameCache[frame:GetName()][TRANSITION_CANCELLED] = true;
    end
end

local function MoveIterate(frame)
    local frameCacheRef = frameCache[frame:GetName()];
    if frameCacheRef[TRANSITION_CANCELLED] then
        q:ResetFrame(frame);
        return;
    elseif not frameCacheRef[TRANSITION_IN_PROGRESS] then
        return;
    end

    local targetLeft = frameCacheRef[TRANSITION_TARGET_LEFT];
    local targetBottom = frameCacheRef[TRANSITION_TARGET_BOTTOM];
    local speed = frameCacheRef[TRANSITION_SPEED];

    -- get current position
    local numPoints, currentPoints = frame:GetNumPoints(), {};
    for i = 1, numPoints do
        local point, relativeTo, relativePoint, ox, oy = frame:GetPoint(i);
        tinsert(currentPoints, { point, relativeTo, relativePoint, ox, oy });
    end

    local done = true;
    frame:ClearAllPoints();
    for i = 1, numPoints do
        -- calculate distance
        local point, relativeTo, relativePoint, ox, oy = unpack(currentPoints[i]);
        local dx, dy = targetLeft - ox, targetBottom - oy;

        -- if distance is above threshold, then move ...
        if math.abs(dx) > maxD or math.abs(dy) > maxD then
            done = false;
            local tx = ox + dx * (speed / 100);
            local ty = oy + dy * (speed / 100);
            frame:SetPoint(point, relativeTo, relativePoint, tx, ty);
        else
            frame:SetPoint(point, relativeTo, relativePoint, ox, oy);
        end
    end

    -- ... and continue recursively
    if done then
        frameCacheRef[TRANSITION_IN_PROGRESS] = false;
        if frameCacheRef[FINISH_FUNC] then
            frameCacheRef[FINISH_FUNC]();
        end
    else
        C_Timer.After(0.01, function ()
            MoveIterate(frame);
        end);
    end
end

function q:TransitionMove(frame, targetLeft, targetBottom, speed, finishFunc)
    if not CheckFrameName(frame) then
        error("Attempt to transition-move anonymous frame cancelled.");
    end

    local frameCacheRef = AddFrame(frame);
    frameCacheRef[TRANSITION_TARGET_LEFT] = targetLeft;
    frameCacheRef[TRANSITION_TARGET_BOTTOM] = targetBottom;
    frameCacheRef[TRANSITION_SPEED] = speed or defaultSpeed;
    frameCacheRef[FINISH_FUNC] = finishFunc;

    if not frameCacheRef[TRANSITION_IN_PROGRESS] then
        frameCacheRef[TRANSITION_IN_PROGRESS] = true;
        MoveIterate(frame);
    end
end